/*
	For best results in visual layout while viewing this file, set
	tab stops to every 4 columns.
*/

/*
	pcmail.c

	copyright (C) 1987 Stuart Lynne

	Copying and use of this program are controlled by the terms of the
	Free Software Foundations GNU Emacs General Public License.

	version 0.1	March 31/1987
	version 0.2 May/88		<skl@van-bc.UUCP>

pcmail

	pcmail address1 address2 ... < the.message

description

	An RFC822 compatible (hopefully) mail delivery system for pc's.

	Designed for delivering mail on a pc based system. It will put
	local mail (ie, not @ or ! in address) into files in the default
	mail directory.

	If remote it will put into an outgoing mailbag in the default mail
	directory. Performs a simple bundling of mail messages into one
	file with arguments prepended as To: arg header lines. And adds a
	Message-Lines: header which gives the number of lines in the
	content part of the message (after the first blank line).

		pcmail john jack jill@xyz.uucp < afile

		To: john,
			jack
			jill@xyz.uucp
		X-Message-Lines: xxxx
		Content-Length: yyyy

		...
		...
		...

	Content-Length: is used without X- prepended to be compatible with AT&T
	Mail bundles. This is not 822 compatible per se, but is allowed.

	It also adds the From_, From: and Date: lines. Other RFC 882 header lines
	such as a Subject: line may be inserted by placing them at the beginning
	of the message.

	A Unix version should lock the /usr/mail/mailbag file.

	Another program called rpcmail will unbundle the files created by
	pcmail and deliver each message to the local rmail. So conceptually

		pcmail ..... < ...; pcmail .... < ...) | sz -> rz | rpcmail

	would deliver remote messages intact.

environment variables

	The following evironment variables are used:

	MAILBOX		current user's mailbox	"stuart"
	NAME		current user's name		"Stuart Lynne"
	DOMAIN		domain of this machine	"slmac.wimsey.ca"
	MAILDIR		where is mail kept		"mpw:mail"

compiling

	Compiled by itself it will operate as a standalone program. If the
	compiler option:

		-DNOMAIN

	is used, it will compile as a routine:

		pcmail(argc, argv)
		int	argc;
		char *argv[];

	and can be used internally in other programs.

Customization

	PCMAIL	bundle remote mail in a outgoing mailbag.
	RMAIL	behave like rmail, other behave like lmail.
*/

#include <stdio.h>

#include "lib.h"
#include "host.h"

#define	PCMAILVER	"UUPC/pcmail 1.0095"
#define	UAVERSION	"UUPC/mail 1.095"

#define	FORWARD		"Forward to "

static FILE	*mailfile, *tempfile;
static char	buf[BUFSIZ];
static char	tfilename[100];

#ifdef PCMAIL
static long	lines = 0, bytes = 0;
#endif

#ifdef NOMAIN

#ifdef RMAIL
#define	main	rmail
#else
#define	main	lmail
#endif /*RMAIL*/

#define	exit	return

extern int debuglevel;

#else

int	debuglevel;

#endif	/*NOMAIN*/

#ifndef RMAIL
static char	*Subject = nil(char);
#endif /*!RMAIL*/

static long	getseq();
static int sendone();
static char *arpadate();

/* determine if a given address is remote or local */
#define remote_address(cp)	(strpbrk((cp), "!@") != nil(char))


main(argc, argv)
char *argv[];
{
	register int header;
	int argcount;
	char **argvec;
	char remotes[BUFSIZ];
	char *mcurdir;

#ifndef NOMAIN

	/* get environment var's */
	HOSTINIT;
	loadenv();

	if (argc <= 1) {
		fprintf(stderr, "pcmail usage: pcmail addresses < message\n");
		exit(1);
	}
	debuglevel = 1;

#endif /*NOMAIN*/

	if (debuglevel > 5) {
		printmsg(5, "pcmail: argc %d ", argc);
		argcount = argc;
		argvec = argv;
		while (argcount--)
			printmsg(5, " \"%s\"", *argvec++);
		printmsg(5, "date=%s", arpadate());
	}

	mcurdir = getcwd(nil(char), 128);
	CHDIR(spooldir);

	{	/* gemerate a temporary file name */
    	char miscbuff[100];
		sprintf(miscbuff, TFILENAME, getseq());
		mkfilename(tfilename, tempdir, miscbuff);
	}

	printmsg(5, "pcmail: opening %s", tfilename);
	if ((tempfile = FOPEN(tfilename, "w", BINARY)) == nil(FILE)) {
		fprintf(stderr, "pcmail: can't open %s\n", tfilename);
		exit(1);
	}

	/* copy stdin to tempfile, counting content lines and bytes */
	header = TRUE;
	while (fgets(buf, BUFSIZ, stdin) != nil(char)) {
		if (header) {
			if (*buf == '\n')
				header = FALSE;
			else if (strchr(buf, ':') == nil(char)) {
				header = FALSE;
				fprintf(tempfile, "\n");
			}
		} else {
#ifdef PCMAIL
			lines++;
			bytes += strlen(buf);
#endif
		}
		fputs(buf, tempfile);
	}

	fclose(tempfile);

#ifdef PCMAIL
	printmsg(4, "pcmail: stdin copied to tmp (%ld bytes, %ld lines)",
		bytes, lines);
#endif
	printmsg(4, "pcmail: argc=%d", argc);

	/* loop on args, copying to appropriate postbox, doing remote only once
	   remote checking is done empirically, could be better */

#ifndef RMAIL
	if (equal(argv[1], "-s")) {
		argv += 2;
		argc -= 2;
		if (argc == 0)
			return -1;
		Subject = strdup(*argv);
	}
#endif

	argcount = argc;
	argvec = argv;
	*remotes = '\0';

	while (--argcount > 0) {
		argvec++;
		printmsg(5, "pcmail: argv[%d]=%s", argcount, *argvec);

		if (remote_address(*argvec)) {
        	int rlen;

			printmsg(5, "pcmail: remote address");

			/* can we cram one more address on line? */
			rlen = strlen(remotes);
			if ((rlen > 0) && (rlen + 1 + strlen(*argvec) > BUFSIZ)) {
				/* no, too bad, dump it then */
				sendone(argc, argv, remotes + 1, TRUE);	/* remote delivery */
				*remotes = '\0';
			}

			/* add *arvgvec to list of remotes */
			strcat(strcat(remotes, " "), *argvec);

		} else {
			printmsg(5, "pcmail: calling sendone \"%s\"", *argvec);
			sendone(argc, argv, *argvec, FALSE);	/* local delivery */
		}

	}

	/* dump any remotes if necessary */
	if (strlen(remotes) > 0)
		sendone(argc, argv, remotes + 1, TRUE);	/* remote delivery */

#ifndef RMAIL
	if (E_filesent != nil(char)) {
    	char mailsent[100];
		mkfilename(mailsent, homedir, E_filesent);
		printmsg(4, "pcmail: copyfile=%s", mailsent);
		sendone(argc, argv, mailsent, FALSE);
	}
#endif

	unlink(tfilename);
	CHDIR(mcurdir);

	exit(0);

} /*main*/



/*
   sendone - copies file plus headers to appropriate mailbox

   NB. we do headers here to allow flexibility later, for example,
   in being able to do Bcc, per host service processing etc.
*/

static int sendone(argc, argv, address, remote)
char **argv;
int argc;
char *address;
int remote;		/* remote mail? */
{
	char *timestr;
	char icfile[32];	/* local C file		*/
	char ixfile[32];	/* local X file		*/
	char hixfile[32];	/* host's ixfile	*/
	char idfile[32];	/* local D file		*/
	char hidfile[32];	/* host's idfile	*/
	char rxfile[32];	/* remote X file	*/
	char rdfile[32];	/* remote D file	*/
	char tmfile[32];	/* temporary storage	*/
#ifndef RMAIL
	long tloc;
#endif

	if (remote) {
		static char	*spool_fmt = "%c.%.7s%04.4ld";	/* spool file name */
		long seqno;
		/* sprintf all required file names */
		sprintf(tmfile, spool_fmt, 'C', mailserv, seqno = getseq());
		importpath(icfile, tmfile);
		sprintf(idfile, spool_fmt, 'D', mailserv, seqno);
		importpath(hidfile, idfile);
		sprintf(ixfile, spool_fmt, 'D', nodename, seqno);
		importpath(hixfile, ixfile);
		sprintf(rdfile, spool_fmt, 'D', nodename, seqno);
		sprintf(rxfile, spool_fmt, 'X', nodename, seqno);
	} else
		/* postbox file name */
		if (strchr(address, SEPCHAR) == nil(char))
			mkfilename(hidfile, maildir, address);
		else
			strcpy(hidfile, address);

	printmsg(5, "sendone: hidfile=%s", hidfile);

	if (!remote) {
		/* check for forwarding */
		printmsg(5, "sendone: checking for forwarding");
		if ((mailfile = FOPEN(hidfile, "r", TEXT)) != nil(FILE)) {
			char *cp = fgets(buf, BUFSIZ, mailfile);
			fclose(mailfile);
			if ((cp != nil(char)) && (equaln(buf, FORWARD, strlen(FORWARD)))) {
				if (*(cp += strlen(cp) - 1) == '\n')
					*cp = '\0';		/* zap trailing NL */
				cp = buf + strlen(FORWARD);
				return sendone(argc, argv, cp, remote_address(cp));
			}
		}
		/* open mailfile */
		if ((mailfile = FOPEN(hidfile, "a", TEXT)) == nil(FILE)) {
			printmsg(0, "sendone: cannot append to %s", hidfile);
			return FALSE;
		}
	} else /*remote*/
		if ((mailfile = FOPEN(hidfile, "w", BINARY)) == nil(FILE)) {
			printmsg(0, "sendone: cannot write to %s", hidfile);
			return FALSE;
		}

	printmsg(5, "sendone: writing to mailfile");

	timestr = strdup(arpadate());

#ifdef RMAIL
	fprintf(mailfile, "From uucp %s\n", timestr);
#else
	fprintf(mailfile, "From %s %s", mailbox, timestr);
	if (remote)
		fprintf(mailfile, " remote from %s", nodename);
	fputc('\n', mailfile);
#endif

	fprintf(mailfile, "Received: by %s (%s) with UUCP; %s\n",
		domain, PCMAILVER, timestr);

#ifndef RMAIL
	tloc = time(nil(long));

	fprintf(mailfile, "Date: %s\n", timestr);
	fprintf(mailfile, "From: %s <%s@%s>\n", name, mailbox, domain);
	fprintf(mailfile, "Message-ID: <%lx@%s>\n", tloc, domain);
	fprintf(mailfile, "X-Mailer: %s\n", UAVERSION);

	/* generate the To: line in a nice format */
	argc--;
	fprintf(mailfile, "To: %s", *++argv);
	while (--argc > 0)
		fprintf(mailfile, ",\n    %s", *++argv);
	fputc('\n', mailfile);

	if (Subject != nil(char))
		fprintf(mailfile, "Subject: %s\n", Subject);

#ifdef PCMAIL
	/* add Message-Lines: and Content-Length: headers */
	fprintf(mailfile, "X-Message-Lines: %ld\n", lines);
	fprintf(mailfile, "Content-Length: %ld\n", bytes);
#endif

#endif /*!RMAIL*/

	/* copy tempfile to mailbox file */
	printmsg(4, "sendone: copying tempfile (%s) to mailfile (%s)",
		tfilename, hidfile);
	if ((tempfile = FOPEN(tfilename, "r", BINARY)) == nil(FILE)) {
		printmsg(0, "sendone: can't re-open %s", tfilename);
		return FALSE;
	}
	while (fgets(buf, BUFSIZ, tempfile) != nil(char)) {
		if (equaln(buf, "From ", 5))
			fputc('>', mailfile);
		fputs(buf, mailfile);
		if (buf[strlen(buf)-1] != '\n')
			fputc('\n', mailfile);
	}

	/* close files */
	fclose(mailfile);
	fclose(tempfile);

	/* if sending to remote (via UUCP), create job control files */
	if (remote) {
		static char *send_cmd = "S %s %s uucp - %s 0666 uucp\n";

		/* create remote X (xqt) file */
		if ((mailfile = FOPEN(hixfile, "w", BINARY)) == nil(FILE)) {
			printmsg(0, "sendone: cannot open X file %s", hixfile);
			return FALSE;
		}
		fprintf(mailfile, "U uucp %s\nF %s\nI %s\nC rmail %s\n",
			nodename, rdfile, rdfile, address);
		fclose(mailfile);

		/* create local C (call) file */
		mailfile = FOPEN(icfile, "w", TEXT);
		if (mailfile == nil(FILE)) {
			printmsg(0, "sendone: cannot open C file %s", icfile);
			return FALSE;
		}
		fprintf(mailfile, send_cmd, idfile, rdfile, idfile);
		fprintf(mailfile, send_cmd, ixfile, rxfile, ixfile);
		fclose(mailfile);

	}

	return TRUE;

} /*sendone*/


static long getseq()
{
	char seqfile[100];
	FILE *seqfile_fp;
	long seq;

	mkfilename(seqfile, confdir, SFILENAME);
	printmsg(4, "getseq: opening %s", seqfile);
	if ((seqfile_fp = FOPEN(seqfile, "r", TEXT)) != nil(FILE)) {
		fscanf(seqfile_fp, "%ld", &seq);
		fclose(seqfile_fp);
	} else {
		printmsg(0, "getseq: can't find %s, creating", seqfile);
		seq = 1;
	};

	/* update sequence number */
	printmsg(5, "getseq: seq#=%ld", seq);

	if ((seqfile_fp = FOPEN(seqfile, "w", TEXT)) != nil(FILE)) {
		fprintf(seqfile_fp, "%ld\n", seq + 1);
		fclose(seqfile_fp);
	}

	return seq;

} /*getseq*/


/*
	arpadate - return the current date/time in RFC 822 format

	ctime() format	'Mon Nov 21 11:31:54 1983\n\0'

	RFC822 format	'Mon, 21 Nov 83 11:31:54 PST\0' or
	RFC822 format	'Mon, 16 May 88 02:13:10 -0700\0'
*/

static char *arpadate()
{
	static char adate[30];
	char *cdate;
	time_t t;

	time(&t);
	cdate = ctime(&t);
	strncpy(adate, cdate, 3);		/* week day	*/
	strcpy(adate + 3, ",");
	strncat(adate, cdate + 7, 3);	/* day		*/
	strncat(adate, cdate + 3, 5);	/* month	*/
	strncat(adate, cdate + 22, 2);	/* year		*/
	strncat(adate, cdate + 10, 10);	/* time		*/
	if (*E_tzname != '\0')
		strncat(adate, E_tzname, 3);	/* time zone */
	else {
		char zone[6];
		int offset = E_tzoffset;
		if (offset < 0) {
			offset = - offset;
			strcat(adate, "-");
		} else
			strcat(adate, "+");
		sprintf(zone, "%02d%02d", (offset / 60) % 12, offset % 60);
		strcat(adate, zone);
	}

	return(adate);

} /*arpadate*/
